@@ -4,7 +4,7 @@
 //!
 //! | Rust type                             | MySQL type(s)                                        |
 //! |---------------------------------------|------------------------------------------------------|
-//! | `bool`                                | TINYINT(1)                                           |
+//! | `bool`                                | TINYINT(1), BOOLEAN                                  |
 //! | `i8`                                  | TINYINT                                              |
 //! | `i16`                                 | SMALLINT                                             |
 //! | `i32`                                 | INT                                                  |
@@ -25,7 +25,7 @@
 //! | Rust type                             | MySQL type(s)                                        |
 //! |---------------------------------------|------------------------------------------------------|
 //! | `chrono::DateTime<Utc>`               | TIMESTAMP                                            |
-//! | `chrono::DateTime<Local>`             | TIMETAMP                                             |
+//! | `chrono::DateTime<Local>`             | TIMESTAMP                                            |
 //! | `chrono::NaiveDateTime`               | DATETIME                                             |
 //! | `chrono::NaiveDate`                   | DATE                                                 |
 //! | `chrono::NaiveTime`                   | TIME                                                 |
@@ -47,6 +47,23 @@
 //! | Rust type                             | MySQL type(s)                                        |
 //! |---------------------------------------|------------------------------------------------------|
 //! | `bigdecimal::BigDecimal`              | DECIMAL                                              |
+//!
+//! ### [`decimal`](https://crates.io/crates/rust_decimal)
+//! Requires the `decimal` Cargo feature flag.
+//!
+//! | Rust type                             | MySQL type(s)                                        |
+//! |---------------------------------------|------------------------------------------------------|
+//! | `rust_decimal::Decimal`               | DECIMAL                                              |
+//!
+//! ### [`uuid`](https://crates.io/crates/uuid)
+//!
+//! Requires the `uuid` Cargo feature flag.
+//!
+//! | Rust type                             | MySQL type(s)                                        |
+//! |---------------------------------------|------------------------------------------------------|
+//! | `uuid::Uuid`                          | BYTE(16), VARCHAR, CHAR, TEXT                        |
+//! | `uuid::adapter::Hyphenated`           | CHAR(36)                                             |
+//!
 //! ### [`json`](https://crates.io/crates/json)
 //!
 //! Requires the `json` Cargo feature flag.
@@ -71,27 +88,17 @@ mod uint;
 #[cfg(feature = "bigdecimal")]
 mod bigdecimal;
 
+#[cfg(feature = "decimal")]
+mod decimal;
+
 #[cfg(feature = "chrono")]
 mod chrono;
 
 #[cfg(feature = "time")]
 mod time;
 
+#[cfg(feature = "uuid")]
+mod uuid;
+
 #[cfg(feature = "json")]
 mod json;
-
-use crate::decode::Decode;
-use crate::mysql::{MySql, MySqlValue};
-
-impl<'de, T> Decode<'de, MySql> for Option<T>
-where
-    T: Decode<'de, MySql>,
-{
-    fn decode(value: MySqlValue<'de>) -> crate::Result<Self> {
-        Ok(if value.get().is_some() {
-            Some(<T as Decode<MySql>>::decode(value)?)
-        } else {
-            None
-        })
-    }
-}
