@@ -1,45 +1,31 @@
+use super::connection::{Floating, Idle, Live};
+use crate::connection::ConnectOptions;
+use crate::connection::Connection;
+use crate::database::Database;
+use crate::error::Error;
+use crate::pool::{deadline_as_timeout, PoolOptions};
+use crossbeam_queue::{ArrayQueue, SegQueue};
+use futures_core::task::{Poll, Waker};
+use futures_util::future;
+use sqlx_rt::{sleep, spawn, timeout};
 use std::cmp;
 use std::mem;
 use std::ptr;
 use std::sync::atomic::{AtomicBool, AtomicU32, Ordering};
 use std::sync::Arc;
+use std::task::Context;
 use std::time::Instant;
 
-use crossbeam_queue::{ArrayQueue, SegQueue};
-use futures_core::task::{Poll, Waker};
-use futures_util::future;
-
-use crate::pool::deadline_as_timeout;
-use crate::runtime::{sleep, spawn, timeout};
-use crate::{
-    connection::{Connect, Connection},
-    error::Error,
-};
-
-use super::connection::{Floating, Idle, Live};
-use super::Options;
-
-pub(crate) struct SharedPool<C> {
-    url: String,
-    idle_conns: ArrayQueue<Idle<C>>,
-    waiters: SegQueue<Waker>,
+pub(crate) struct SharedPool<DB: Database> {
+    pub(super) connect_options: <DB::Connection as Connection>::Options,
+    pub(super) idle_conns: ArrayQueue<Idle<DB>>,
+    waiters: SegQueue<Arc<Waiter>>,
     pub(super) size: AtomicU32,
     is_closed: AtomicBool,
-    options: Options,
+    pub(super) options: PoolOptions<DB>,
 }
 
-impl<C> SharedPool<C>
-where
-    C: Connection,
-{
-    pub fn options(&self) -> &Options {
-        &self.options
-    }
-
-    pub(super) fn url(&self) -> &str {
-        &self.url
-    }
-
+impl<DB: Database> SharedPool<DB> {
     pub(super) fn size(&self) -> u32 {
         self.size.load(Ordering::Acquire)
     }
@@ -55,18 +41,32 @@ where
 
     pub(super) async fn close(&self) {
         self.is_closed.store(true, Ordering::Release);
-        while let Ok(_) = self.idle_conns.pop() {}
         while let Ok(waker) = self.waiters.pop() {
             waker.wake();
         }
+
+        // ensure we wait until the pool is actually closed
+        while self.size() > 0 {
+            let _ = self
+                .idle_conns
+                .pop()
+                .map(|idle| Floating::from_idle(idle, self));
+
+            // yield to avoid starving the executor
+            sqlx_rt::yield_now().await;
+        }
     }
 
     #[inline]
-    pub(super) fn try_acquire(&self) -> Option<Floating<Live<C>>> {
+    pub(super) fn try_acquire(&self) -> Option<Floating<'_, Live<DB>>> {
+        // don't cut in line
+        if self.options.fair && !self.waiters.is_empty() {
+            return None;
+        }
         Some(self.pop_idle()?.into_live())
     }
 
-    fn pop_idle(&self) -> Option<Floating<Idle<C>>> {
+    fn pop_idle(&self) -> Option<Floating<'_, Idle<DB>>> {
         if self.is_closed.load(Ordering::Acquire) {
             return None;
         }
@@ -74,10 +74,18 @@ where
         Some(Floating::from_idle(self.idle_conns.pop().ok()?, self))
     }
 
-    pub(super) fn release(&self, floating: Floating<Live<C>>) {
+    pub(super) fn release(&self, mut floating: Floating<'_, Live<DB>>) {
+        if let Some(test) = &self.options.after_release {
+            if !test(&mut floating.raw) {
+                // drop the connection and do not return to the pool
+                return;
+            }
+        }
+
         self.idle_conns
             .push(floating.into_idle().into_leakable())
             .expect("BUG: connection queue overflow in release()");
+
         if let Ok(waker) = self.waiters.pop() {
             waker.wake();
         }
@@ -85,11 +93,15 @@ where
 
     /// Try to atomically increment the pool size for a new connection.
     ///
-    /// Returns `None` if we are at max_size.
-    fn try_increment_size(&self) -> Option<DecrementSizeGuard> {
+    /// Returns `None` if we are at max_connections or if the pool is closed.
+    pub(super) fn try_increment_size(&self) -> Option<DecrementSizeGuard<'_>> {
+        if self.is_closed() {
+            return None;
+        }
+
         let mut size = self.size();
 
-        while size < self.options.max_size {
+        while size < self.options.max_connections {
             let new_size = self.size.compare_and_swap(size, size + 1, Ordering::AcqRel);
 
             if new_size == size {
@@ -102,73 +114,80 @@ where
         None
     }
 
-    /// Wait for a connection, if either `size` drops below `max_size` so we can
+    /// Wait for a connection, if either `size` drops below `max_connections` so we can
     /// open a new connection, or if an idle connection is returned to the pool.
     ///
     /// Returns an error if `deadline` elapses before we are woken.
-    async fn wait_for_conn(&self, deadline: Instant) -> crate::Result<()> {
-        let mut waker_pushed = false;
+    async fn wait_for_conn(&self, deadline: Instant) -> Result<(), Error> {
+        if self.is_closed() {
+            return Err(Error::PoolClosed);
+        }
+
+        let mut waiter = None;
 
         timeout(
-            deadline_as_timeout::<C::Database>(deadline)?,
+            deadline_as_timeout::<DB>(deadline)?,
             // `poll_fn` gets us easy access to a `Waker` that we can push to our queue
-            future::poll_fn(|ctx| -> Poll<()> {
-                if !waker_pushed {
-                    // only push the waker once
-                    self.waiters.push(ctx.waker().to_owned());
-                    waker_pushed = true;
-                    Poll::Pending
-                } else {
+            future::poll_fn(|cx| -> Poll<()> {
+                let waiter = waiter.get_or_insert_with(|| {
+                    let waiter = Waiter::new(cx);
+                    self.waiters.push(waiter.clone());
+                    waiter
+                });
+
+                if waiter.is_woken() {
                     Poll::Ready(())
+                } else {
+                    Poll::Pending
                 }
             }),
         )
         .await
-        .map_err(|_| crate::Error::PoolTimedOut(None))
+        .map_err(|_| Error::PoolTimedOut)
     }
-}
 
-impl<C> SharedPool<C>
-where
-    C: Connect,
-{
-    pub(super) async fn new_arc(url: &str, options: Options) -> crate::Result<Arc<Self>> {
-        let mut pool = Self {
-            url: url.to_owned(),
-            idle_conns: ArrayQueue::new(options.max_size as usize),
+    pub(super) fn new_arc(
+        options: PoolOptions<DB>,
+        connect_options: <DB::Connection as Connection>::Options,
+    ) -> Arc<Self> {
+        let pool = Self {
+            connect_options,
+            idle_conns: ArrayQueue::new(options.max_connections as usize),
             waiters: SegQueue::new(),
             size: AtomicU32::new(0),
             is_closed: AtomicBool::new(false),
             options,
         };
 
-        pool.init_min_connections().await?;
-
         let pool = Arc::new(pool);
 
         spawn_reaper(&pool);
 
-        Ok(pool)
+        pool
     }
 
-    pub(super) async fn acquire<'s>(&'s self) -> crate::Result<Floating<'s, Live<C>>> {
+    #[allow(clippy::needless_lifetimes)]
+    pub(super) async fn acquire<'s>(&'s self) -> Result<Floating<'s, Live<DB>>, Error> {
         let start = Instant::now();
         let deadline = start + self.options.connect_timeout;
+        let mut waited = !self.options.fair;
 
         // Unless the pool has been closed ...
         while !self.is_closed() {
-            // Attempt to immediately acquire a connection. This will return Some
-            // if there is an idle connection in our channel.
-            if let Ok(conn) = self.idle_conns.pop() {
-                let conn = Floating::from_idle(conn, self);
-                if let Some(live) = check_conn(conn, &self.options).await {
-                    return Ok(live);
+            // Don't cut in line
+            if waited || self.waiters.is_empty() {
+                // Attempt to immediately acquire a connection. This will return Some
+                // if there is an idle connection in our channel.
+                if let Some(conn) = self.pop_idle() {
+                    if let Some(live) = check_conn(conn, &self.options).await {
+                        return Ok(live);
+                    }
                 }
             }
 
             if let Some(guard) = self.try_increment_size() {
                 // pool has slots available; open a new connection
-                match self.connect(deadline, guard).await {
+                match self.connection(deadline, guard).await {
                     Ok(Some(conn)) => return Ok(conn),
                     // [size] is internally decremented on _retry_ and _error_
                     Ok(None) => continue,
@@ -179,89 +198,74 @@ where
             // Wait for a connection to become available (or we are allowed to open a new one)
             // Returns an error if `deadline` passes
             self.wait_for_conn(deadline).await?;
+
+            waited = true;
         }
 
         Err(Error::PoolClosed)
     }
 
-    // takes `&mut self` so this can only be called during init
-    async fn init_min_connections(&mut self) -> crate::Result<()> {
-        for _ in 0..self.options.min_size {
-            let deadline = Instant::now() + self.options.connect_timeout;
-
-            // this guard will prevent us from exceeding `max_size`
-            while let Some(guard) = self.try_increment_size() {
-                // [connect] will raise an error when past deadline
-                // [connect] returns None if its okay to retry
-                if let Some(conn) = self.connect(deadline, guard).await? {
-                    self.idle_conns
-                        .push(conn.into_idle().into_leakable())
-                        .expect("BUG: connection queue overflow in init_min_connections");
-                }
-            }
-        }
-
-        Ok(())
-    }
-
-    async fn connect<'s>(
+    pub(super) async fn connection<'s>(
         &'s self,
         deadline: Instant,
         guard: DecrementSizeGuard<'s>,
-    ) -> crate::Result<Option<Floating<'s, Live<C>>>> {
+    ) -> Result<Option<Floating<'s, Live<DB>>>, Error> {
         if self.is_closed() {
             return Err(Error::PoolClosed);
         }
 
-        let timeout = super::deadline_as_timeout::<C::Database>(deadline)?;
+        let timeout = super::deadline_as_timeout::<DB>(deadline)?;
 
         // result here is `Result<Result<C, Error>, TimeoutError>`
-        match crate::runtime::timeout(timeout, C::connect(&self.url)).await {
+        match sqlx_rt::timeout(timeout, self.connect_options.connect()).await {
             // successfully established connection
-            Ok(Ok(raw)) => Ok(Some(Floating::new_live(raw, guard))),
+            Ok(Ok(mut raw)) => {
+                if let Some(callback) = &self.options.after_connect {
+                    callback(&mut raw).await?;
+                }
+
+                Ok(Some(Floating::new_live(raw, guard)))
+            }
 
             // an IO error while connecting is assumed to be the system starting up
-            Ok(Err(crate::Error::Io(_))) => Ok(None),
+            Ok(Err(Error::Io(_))) => Ok(None),
 
             // TODO: Handle other database "boot period"s
 
             // [postgres] the database system is starting up
             // TODO: Make this check actually check if this is postgres
-            Ok(Err(crate::Error::Database(error))) if error.code() == Some("57P03") => Ok(None),
+            Ok(Err(Error::Database(error))) if error.code().as_deref() == Some("57P03") => Ok(None),
 
             // Any other error while connection should immediately
             // terminate and bubble the error up
             Ok(Err(e)) => Err(e),
 
             // timed out
-            Err(e) => Err(crate::Error::PoolTimedOut(Some(Box::new(e)))),
+            Err(_) => Err(Error::PoolTimedOut),
         }
     }
 }
 
 // NOTE: Function names here are bizzare. Helpful help would be appreciated.
 
-fn is_beyond_lifetime<C>(live: &Live<C>, options: &Options) -> bool {
+fn is_beyond_lifetime<DB: Database>(live: &Live<DB>, options: &PoolOptions<DB>) -> bool {
     // check if connection was within max lifetime (or not set)
     options
         .max_lifetime
         .map_or(false, |max| live.created.elapsed() > max)
 }
 
-fn is_beyond_idle<C>(idle: &Idle<C>, options: &Options) -> bool {
+fn is_beyond_idle<DB: Database>(idle: &Idle<DB>, options: &PoolOptions<DB>) -> bool {
     // if connection wasn't idle too long (or not set)
     options
         .idle_timeout
         .map_or(false, |timeout| idle.since.elapsed() > timeout)
 }
 
-async fn check_conn<'s: 'p, 'p, C>(
-    mut conn: Floating<'s, Idle<C>>,
-    options: &'p Options,
-) -> Option<Floating<'s, Live<C>>>
-where
-    C: Connection,
-{
+async fn check_conn<'s: 'p, 'p, DB: Database>(
+    mut conn: Floating<'s, Idle<DB>>,
+    options: &'p PoolOptions<DB>,
+) -> Option<Floating<'s, Live<DB>>> {
     // If the connection we pulled has expired, close the connection and
     // immediately create a new connection
     if is_beyond_lifetime(&conn, options) {
@@ -269,8 +273,7 @@ where
         // close the connection but don't really care about the result
         let _ = conn.close().await;
         return None;
-    } else if options.test_on_acquire {
-        // TODO: Check on acquire should be a configuration setting
+    } else if options.test_before_acquire {
         // Check that the connection is still live
         if let Err(e) = conn.ping().await {
             // an error here means the other end has hung up or we lost connectivity
@@ -280,6 +283,20 @@ where
             // connection is broken so don't try to close nicely
             return None;
         }
+    } else if let Some(test) = &options.before_acquire {
+        match test(&mut conn.live.raw).await {
+            Ok(false) => {
+                // connection was rejected by user-defined hook
+                return None;
+            }
+
+            Err(error) => {
+                log::info!("in `before_acquire`: {}", error);
+                return None;
+            }
+
+            Ok(true) => {}
+        }
     }
 
     // No need to re-connect; connection is alive or we don't care
@@ -287,10 +304,7 @@ where
 }
 
 /// if `max_lifetime` or `idle_timeout` is set, spawn a task that reaps senescent connections
-fn spawn_reaper<C>(pool: &Arc<SharedPool<C>>)
-where
-    C: Connection,
-{
+fn spawn_reaper<DB: Database>(pool: &Arc<SharedPool<DB>>) {
     let period = match (pool.options.max_lifetime, pool.options.idle_timeout) {
         (Some(it), None) | (None, Some(it)) => it,
 
@@ -304,7 +318,7 @@ where
     spawn(async move {
         while !pool.is_closed.load(Ordering::Acquire) {
             // reap at most the current size minus the minimum idle
-            let max_reaped = pool.size().saturating_sub(pool.options.min_size);
+            let max_reaped = pool.size().saturating_sub(pool.options.min_connections);
 
             // collect connections to reap
             let (reap, keep) = (0..max_reaped)
@@ -336,12 +350,12 @@ where
 /// (where the pool thinks it has more connections than it does).
 pub(in crate::pool) struct DecrementSizeGuard<'a> {
     size: &'a AtomicU32,
-    waiters: &'a SegQueue<Waker>,
+    waiters: &'a SegQueue<Arc<Waiter>>,
     dropped: bool,
 }
 
 impl<'a> DecrementSizeGuard<'a> {
-    pub fn new<C>(pool: &'a SharedPool<C>) -> Self {
+    pub fn new<DB: Database>(pool: &'a SharedPool<DB>) -> Self {
         Self {
             size: &pool.size,
             waiters: &pool.waiters,
@@ -350,7 +364,7 @@ impl<'a> DecrementSizeGuard<'a> {
     }
 
     /// Return `true` if the internal references point to the same fields in `SharedPool`.
-    pub fn same_pool<C>(&self, pool: &'a SharedPool<C>) -> bool {
+    pub fn same_pool<DB: Database>(&self, pool: &'a SharedPool<DB>) -> bool {
         ptr::eq(self.size, &pool.size) && ptr::eq(self.waiters, &pool.waiters)
     }
 
@@ -369,3 +383,26 @@ impl Drop for DecrementSizeGuard<'_> {
         }
     }
 }
+
+struct Waiter {
+    woken: AtomicBool,
+    waker: Waker,
+}
+
+impl Waiter {
+    fn new(cx: &mut Context<'_>) -> Arc<Self> {
+        Arc::new(Self {
+            woken: AtomicBool::new(false),
+            waker: cx.waker().clone(),
+        })
+    }
+
+    fn wake(&self) {
+        self.woken.store(true, Ordering::Release);
+        self.waker.wake_by_ref();
+    }
+
+    fn is_woken(&self) -> bool {
+        self.woken.load(Ordering::Acquire)
+    }
+}
